﻿namespace Menees
{
	#region Using Directives

	using System;
	using System.Collections.Generic;
	using System.Linq;
	using System.Security.Principal;
	using System.Text;
	using System.Threading.Tasks;

	#endregion

	public static partial class ApplicationInfo
	{
		#region Public Properties

		/// <summary>
		/// Gets whether the current user is running in the Windows "Administrator" role.
		/// </summary>
		public static bool IsUserRunningAsAdministrator
		{
			get
			{
				using (WindowsIdentity currentIdentity = WindowsIdentity.GetCurrent())
				{
					WindowsPrincipal currentPrincipal = new WindowsPrincipal(currentIdentity);
					bool result = currentPrincipal.IsInRole(WindowsBuiltInRole.Administrator);
					return result;
				}
			}
		}

		#endregion

		#region Public Methods

		/// <summary>
		/// Creates a hierarchical store for loading and saving user-level settings for the current application.
		/// </summary>
		/// <returns>A new settings store.</returns>
		public static ISettingsStore CreateUserSettingsStore()
		{
			ISettingsStore result;

			// The registry has the advantage that it truly isolates each user's
			// settings from each other, so one (non-admin) user can't see or
			// edit another user's settings on disk.  However, the registry is also
			// shared between multiple instances of the same app on one machine
			// (e.g., if you're testing ClientA and ClientB on one box), and the
			// settings for the app hang around in the registry even if you delete
			// the application's folder.
			//
			// The file store has the advantage that by default everything is located
			// in the current application directory.  However, if the user doesn't
			// have permissions to write to the application directory, then it falls back
			// to using a user's shared folder (e.g., AppData\Local or %Temp%),
			// which means it would lose the side-by-side isolation ability.
			//
			// We'll default to the file store because it's the best for most purposes.
			if (Properties.Settings.Default.StoreUserSettingsInRegistry)
			{
				result = new RegistrySettingsStore();
			}
			else
			{
				result = new FileSettingsStore();
			}

			return result;
		}

		#endregion
	}
}
